"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var turbocolor_1 = __importDefault(require("turbocolor"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
function batchWarnings() {
    var allWarnings = new Map();
    var count = 0;
    return {
        get count() {
            return count;
        },
        add: function (warning) {
            if (typeof warning === 'string') {
                warning = { code: 'UNKNOWN', message: warning };
            }
            if (warning.code in immediateHandlers) {
                immediateHandlers[warning.code](warning);
                return;
            }
            if (!allWarnings.has(warning.code))
                allWarnings.set(warning.code, []);
            allWarnings.get(warning.code).push(warning);
            count += 1;
        },
        flush: function () {
            if (count === 0)
                return;
            var codes = Array.from(allWarnings.keys()).sort(function (a, b) {
                if (deferredHandlers[a] && deferredHandlers[b]) {
                    return deferredHandlers[a].priority - deferredHandlers[b].priority;
                }
                if (deferredHandlers[a])
                    return -1;
                if (deferredHandlers[b])
                    return 1;
                return allWarnings.get(b).length - allWarnings.get(a).length;
            });
            codes.forEach(function (code) {
                var handler = deferredHandlers[code];
                var warnings = allWarnings.get(code);
                if (handler) {
                    handler.fn(warnings);
                }
                else {
                    warnings.forEach(function (warning) {
                        logging_1.stderr(turbocolor_1.default.bold.yellow('(!)') + " " + turbocolor_1.default.bold.yellow(warning.message));
                        if (warning.url)
                            info(warning.url);
                        var id = (warning.loc && warning.loc.file) || warning.id;
                        if (id) {
                            var loc = warning.loc
                                ? relativeId_1.default(id) + ": (" + warning.loc.line + ":" + warning.loc.column + ")"
                                : relativeId_1.default(id);
                            logging_1.stderr(turbocolor_1.default.bold(relativeId_1.default(loc)));
                        }
                        if (warning.frame)
                            info(warning.frame);
                    });
                }
            });
            allWarnings = new Map();
            count = 0;
        }
    };
}
exports.default = batchWarnings;
var immediateHandlers = {
    UNKNOWN_OPTION: function (warning) {
        title("You have passed an unrecognized option");
        logging_1.stderr(warning.message);
    },
    DEPRECATED_OPTIONS: function (warning) {
        title("Some options have been renamed");
        info("https://gist.github.com/Rich-Harris/d472c50732dab03efeb37472b08a3f32");
        warning.deprecations.forEach(function (option) {
            logging_1.stderr(turbocolor_1.default.bold(option.old) + " is now " + option.new);
        });
    },
    MISSING_NODE_BUILTINS: function (warning) {
        title("Missing shims for Node.js built-ins");
        var detail = warning.modules.length === 1
            ? "'" + warning.modules[0] + "'"
            : warning.modules
                .slice(0, -1)
                .map(function (name) { return "'" + name + "'"; })
                .join(', ') + " and '" + warning.modules.slice(-1) + "'";
        logging_1.stderr("Creating a browser bundle that depends on " + detail + ". You might need to include https://www.npmjs.com/package/rollup-plugin-node-builtins");
    },
    MIXED_EXPORTS: function () {
        title('Mixing named and default exports');
        logging_1.stderr("Consumers of your bundle will have to use bundle['default'] to access the default export, which may not be what you want. Use `output.exports: 'named'` to disable this warning");
    },
    EMPTY_BUNDLE: function () {
        title("Generated an empty bundle");
    }
};
// TODO select sensible priorities
var deferredHandlers = {
    UNUSED_EXTERNAL_IMPORT: {
        priority: 1,
        fn: function (warnings) {
            title('Unused external imports');
            warnings.forEach(function (warning) {
                logging_1.stderr(warning.names + " imported from external module '" + warning.source + "' but never used");
            });
        }
    },
    UNRESOLVED_IMPORT: {
        priority: 1,
        fn: function (warnings) {
            title('Unresolved dependencies');
            info('https://rollupjs.org/guide/en#warning-treating-module-as-external-dependency');
            var dependencies = new Map();
            warnings.forEach(function (warning) {
                if (!dependencies.has(warning.source))
                    dependencies.set(warning.source, []);
                dependencies.get(warning.source).push(warning.importer);
            });
            Array.from(dependencies.keys()).forEach(function (dependency) {
                var importers = dependencies.get(dependency);
                logging_1.stderr(turbocolor_1.default.bold(dependency) + " (imported by " + importers.join(', ') + ")");
            });
        }
    },
    MISSING_EXPORT: {
        priority: 1,
        fn: function (warnings) {
            title('Missing exports');
            info('https://rollupjs.org/guide/en#error-name-is-not-exported-by-module-');
            warnings.forEach(function (warning) {
                logging_1.stderr(turbocolor_1.default.bold(warning.importer));
                logging_1.stderr(warning.missing + " is not exported by " + warning.exporter);
                logging_1.stderr(turbocolor_1.default.gray(warning.frame));
            });
        }
    },
    THIS_IS_UNDEFINED: {
        priority: 1,
        fn: function (warnings) {
            title('`this` has been rewritten to `undefined`');
            info('https://rollupjs.org/guide/en#error-this-is-undefined');
            showTruncatedWarnings(warnings);
        }
    },
    EVAL: {
        priority: 1,
        fn: function (warnings) {
            title('Use of eval is strongly discouraged');
            info('https://rollupjs.org/guide/en#avoiding-eval');
            showTruncatedWarnings(warnings);
        }
    },
    NON_EXISTENT_EXPORT: {
        priority: 1,
        fn: function (warnings) {
            title("Import of non-existent " + (warnings.length > 1 ? 'exports' : 'export'));
            showTruncatedWarnings(warnings);
        }
    },
    NAMESPACE_CONFLICT: {
        priority: 1,
        fn: function (warnings) {
            title("Conflicting re-exports");
            warnings.forEach(function (warning) {
                logging_1.stderr(turbocolor_1.default.bold(relativeId_1.default(warning.reexporter)) + " re-exports '" + warning.name + "' from both " + relativeId_1.default(warning.sources[0]) + " and " + relativeId_1.default(warning.sources[1]) + " (will be ignored)");
            });
        }
    },
    MISSING_GLOBAL_NAME: {
        priority: 1,
        fn: function (warnings) {
            title("Missing global variable " + (warnings.length > 1 ? 'names' : 'name'));
            logging_1.stderr("Use output.globals to specify browser global variable names corresponding to external modules");
            warnings.forEach(function (warning) {
                logging_1.stderr(turbocolor_1.default.bold(warning.source) + " (guessing '" + warning.guess + "')");
            });
        }
    },
    SOURCEMAP_BROKEN: {
        priority: 1,
        fn: function (warnings) {
            title("Broken sourcemap");
            info('https://rollupjs.org/guide/en#warning-sourcemap-is-likely-to-be-incorrect');
            var plugins = Array.from(new Set(warnings.map(function (w) { return w.plugin; }).filter(Boolean)));
            var detail = plugins.length === 0
                ? ''
                : plugins.length > 1
                    ? " (such as " + plugins
                        .slice(0, -1)
                        .map(function (p) { return "'" + p + "'"; })
                        .join(', ') + " and '" + plugins.slice(-1) + "')"
                    : " (such as '" + plugins[0] + "')";
            logging_1.stderr("Plugins that transform code" + detail + " should generate accompanying sourcemaps");
        }
    },
    PLUGIN_WARNING: {
        priority: 1,
        fn: function (warnings) {
            var nestedByPlugin = nest(warnings, 'plugin');
            nestedByPlugin.forEach(function (_a) {
                var plugin = _a.key, items = _a.items;
                var nestedByMessage = nest(items, 'message');
                var lastUrl;
                nestedByMessage.forEach(function (_a) {
                    var message = _a.key, items = _a.items;
                    title(plugin + " plugin: " + message);
                    items.forEach(function (warning) {
                        if (warning.url !== lastUrl)
                            info((lastUrl = warning.url));
                        var loc = warning.loc
                            ? relativeId_1.default(warning.id) + ": (" + warning.loc.line + ":" + warning.loc.column + ")"
                            : relativeId_1.default(warning.id);
                        logging_1.stderr(turbocolor_1.default.bold(relativeId_1.default(loc)));
                        if (warning.frame)
                            info(warning.frame);
                    });
                });
            });
        }
    }
};
function title(str) {
    logging_1.stderr(turbocolor_1.default.bold.yellow('(!)') + " " + turbocolor_1.default.bold.yellow(str));
}
function info(url) {
    logging_1.stderr(turbocolor_1.default.gray(url));
}
function nest(array, prop) {
    var nested = [];
    var lookup = new Map();
    array.forEach(function (item) {
        var key = item[prop];
        if (!lookup.has(key)) {
            lookup.set(key, {
                key: key,
                items: []
            });
            nested.push(lookup.get(key));
        }
        lookup.get(key).items.push(item);
    });
    return nested;
}
function showTruncatedWarnings(warnings) {
    var nestedByModule = nest(warnings, 'id');
    var sliced = nestedByModule.length > 5 ? nestedByModule.slice(0, 3) : nestedByModule;
    sliced.forEach(function (_a) {
        var id = _a.key, items = _a.items;
        logging_1.stderr(turbocolor_1.default.bold(relativeId_1.default(id)));
        logging_1.stderr(turbocolor_1.default.gray(items[0].frame));
        if (items.length > 1) {
            logging_1.stderr("...and " + (items.length - 1) + " other " + (items.length > 2 ? 'occurrences' : 'occurrence'));
        }
    });
    if (nestedByModule.length > sliced.length) {
        logging_1.stderr("\n...and " + (nestedByModule.length - sliced.length) + " other files");
    }
}
